"""
This module provides tools for working with RedVox API M location data and metadata.
"""

import enum

import redvox.api1000.common.typing
import redvox.api1000.common.common as common

from typing import List

import redvox.api1000.common.generic

from redvox.api1000.common.decorators import wrap_enum
from redvox.api1000.common.generic import ProtoBase, ProtoRepeatedMessage
from redvox.api1000.proto.redvox_api_m_pb2 import RedvoxPacketM


# noinspection Mypy
@wrap_enum(RedvoxPacketM.Sensors.Location.LocationProvider)
class LocationProvider(enum.Enum):
    """
    An enum that represents valid location providers
    """
    UNKNOWN: int = 0
    NONE: int = 1
    USER: int = 2
    GPS: int = 3
    NETWORK: int = 4


# noinspection Mypy
@wrap_enum(RedvoxPacketM.Sensors.Location.BestLocation.LocationScoreMethod)
class LocationScoreMethod(enum.Enum):
    """
    An enum that represents location scoring methods
    """
    UNKNOWN_METHOD: int = 0


class BestTimestamp(
    redvox.api1000.common.generic.ProtoBase[RedvoxPacketM.Sensors.Location.BestLocation.BestTimestamp]
):
    """
    This class encapsulates the best timestamps associated with the best location estimates.
    """
    def __init__(self, proto: RedvoxPacketM.Sensors.Location.BestLocation.BestTimestamp):
        super().__init__(proto)

    def get_unit(self) -> common.Unit:
        """
        :return: The unit of the timestamps.
        """
        # noinspection Mypy
        return common.Unit.from_proto(self._proto.unit)

    def set_default_unit(self) -> 'BestTimestamp':
        """
        Sets the default timestamp unit.
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        return self.set_unit(common.Unit.MICROSECONDS_SINCE_UNIX_EPOCH)

    def set_unit(self, unit: common.Unit) -> 'BestTimestamp':
        """
        Sets the unit of the best timestamps.
        :param unit: Unit to set.
        :return: A modified instance of self.
        """
        redvox.api1000.common.typing.check_type(unit, [common.Unit])
        # noinspection Mypy
        self._proto.unit = unit.into_proto()
        return self

    def get_mach(self) -> float:
        """
        :return: The best machine timestamp
        """
        return self._proto.mach

    def set_mach(self, mach: float) -> 'BestTimestamp':
        """
        Sets the best machine timestamp.
        :param mach: Timestamp to set.
        :return: A modified instance of self.
        """
        redvox.api1000.common.typing.check_type(mach, [int, float])
        self._proto.mach = mach
        return self

    def get_gps(self) -> float:
        """
        :return: The best GPS timestamp.
        """
        return self._proto.gps

    def set_gps(self, gps: float) -> 'BestTimestamp':
        """
        Sets the best GPS provided timestamp
        :param gps: Timestamp to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(gps, [int, float])
        self._proto.gps = gps
        return self


class BestLocation(ProtoBase[RedvoxPacketM.Sensors.Location.BestLocation]):
    """
    Encapsulates the best location estimates
    """
    def __init__(self, proto: RedvoxPacketM.Sensors.Location.BestLocation):
        super().__init__(proto)
        self._latitude_longitude_timestamp: BestTimestamp = BestTimestamp(proto.latitude_longitude_timestamp)
        self._altitude_timestamp: BestTimestamp = BestTimestamp(proto.altitude_timestamp)
        self._speed_timestamp: BestTimestamp = BestTimestamp(proto.speed_timestamp)
        self._bearing_timestamp: BestTimestamp = BestTimestamp(proto.bearing_timestamp)

    def get_latitude_longitude_timestamp(self) -> BestTimestamp:
        """
        :return: Best timestamps associated with best latitude and longitude
        """
        return self._latitude_longitude_timestamp

    def get_altitude_timestamp(self) -> BestTimestamp:
        """
        :return: Best timestamps associated with best altitude
        """
        return self._altitude_timestamp

    def get_speed_timestamp(self) -> BestTimestamp:
        """
        :return: Best timestamps associated with best speed
        """
        return self._speed_timestamp

    def get_bearing_timestamp(self) -> BestTimestamp:
        """
        :return: Best timestamps associated with best bearing
        """
        return self._bearing_timestamp

    def get_latitude_longitude_unit(self) -> common.Unit:
        """
        :return: Unit associated with latitude, longitude
        """
        # noinspection Mypy
        return common.Unit.from_proto(self._proto.latitude_longitude_unit)

    def set_latitude_longitude_unit(self, unit: common.Unit) -> 'BestLocation':
        """
        Sets the unit associated with latitude and longitude
        :param unit: Unit to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(unit, [common.Unit])
        # noinspection Mypy
        self._proto.latitude_longitude_unit = unit.into_proto()
        return self

    def get_altitude_unit(self) -> common.Unit:
        """
        :return: Unit associated with altitude
        """
        # noinspection Mypy
        return common.Unit.from_proto(self._proto.altitude_unit)

    def set_altitude_unit(self, unit: common.Unit) -> 'BestLocation':
        """
        Sets the unit associated with altitude
        :param unit: Unit to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(unit, [common.Unit])
        # noinspection Mypy
        self._proto.altitude_unit = unit.into_proto()
        return self

    def get_speed_unit(self) -> common.Unit:
        """
        :return: Unit associated with altitude
        """
        # noinspection Mypy
        return common.Unit.from_proto(self._proto.speed_unit)

    def set_speed_unit(self, unit: common.Unit) -> 'BestLocation':
        """
        Sets the unit associated with speed
        :param unit: Unit to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(unit, [common.Unit])
        # noinspection Mypy
        self._proto.speed_unit = unit.into_proto()
        return self

    def get_bearing_unit(self) -> common.Unit:
        """
        :return: Unit associated with bearing
        """
        # noinspection Mypy
        return common.Unit.from_proto(self._proto.bearing_unit)

    def set_bearing_unit(self, unit: common.Unit) -> 'BestLocation':
        """
        Sets the unit associated with speed
        :param unit: Unit to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(unit, [common.Unit])
        # noinspection Mypy
        self._proto.bearing_unit = unit.into_proto()
        return self

    def get_vertical_accuracy_unit(self) -> common.Unit:
        """
        :return: Unit associated with vertical accuracy
        """
        # noinspection Mypy
        return common.Unit.from_proto(self._proto.vertical_accuracy_unit)

    def set_vertical_accuracy_unit(self, unit: common.Unit) -> 'BestLocation':
        """
        Sets the unit associated with vertical accuracy
        :param unit: Unit to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(unit, [common.Unit])
        # noinspection Mypy
        self._proto.vertical_accuracy_unit = unit.into_proto()
        return self

    def get_horizontal_accuracy_unit(self) -> common.Unit:
        """
        :return: Unit associated with horizontal accuracy
        """
        # noinspection Mypy
        return common.Unit.from_proto(self._proto.horizontal_accuracy_unit)

    def set_horizontal_accuracy_unit(self, unit: common.Unit) -> 'BestLocation':
        """
        Sets the unit associated with horizontal accuracy
        :param unit: Unit to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(unit, [common.Unit])
        # noinspection Mypy
        self._proto.horizontal_accuracy_unit = unit.into_proto()
        return self

    def get_speed_accuracy_unit(self) -> common.Unit:
        """
        :return: Unit associated with speed accuracy
        """
        # noinspection Mypy
        return common.Unit.from_proto(self._proto.speed_accuracy_unit)

    def set_speed_accuracy_unit(self, unit: common.Unit) -> 'BestLocation':
        """
        Sets the unit associated with speed accuracy
        :param unit: Unit to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(unit, [common.Unit])
        # noinspection Mypy
        self._proto.speed_accuracy_unit = unit.into_proto()
        return self

    def get_bearing_accuracy_unit(self) -> common.Unit:
        """
        :return: Unit associated with bearing accuracy
        """
        # noinspection Mypy
        return common.Unit.from_proto(self._proto.bearing_accuracy_unit)

    def set_bearing_accuracy_unit(self, unit: common.Unit) -> 'BestLocation':
        """
        Sets the unit associated with bearing accuracy
        :param unit: Unit to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(unit, [common.Unit])
        # noinspection Mypy
        self._proto.bearing_accuracy_unit = unit.into_proto()
        return self

    def get_latitude(self) -> float:
        """
        :return: The best latitude
        """
        return self._proto.latitude

    def set_latitude(self, latitude: float) -> 'BestLocation':
        """
        Sets the best latitude
        :param latitude: Latitude to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(latitude, [int, float])
        self._proto.latitude = latitude
        return self

    def get_longitude(self) -> float:
        """
        :return: The best longitude
        """
        return self._proto.longitude

    def set_longitude(self, longitude: float) -> 'BestLocation':
        """
        Sets the best longitude
        :param longitude: Longitude to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(longitude, [int, float])
        self._proto.longitude = longitude
        return self

    def get_altitude(self) -> float:
        """
        :return: The best altitude
        """
        return self._proto.altitude

    def set_altitude(self, altitude: float) -> 'BestLocation':
        """
        Sets the best altitude
        :param altitude: Altitude to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(altitude, [int, float])
        self._proto.altitude = altitude
        return self

    def get_speed(self) -> float:
        """
        :return: The best speed
        """
        return self._proto.speed

    def set_speed(self, speed: float) -> 'BestLocation':
        """
        Sets the best speed
        :param speed: Speed to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(speed, [int, float])
        self._proto.speed = speed
        return self

    def get_bearing(self) -> float:
        """
        :return: The best bearing
        """
        return self._proto.bearing

    def set_bearing(self, bearing: float) -> 'BestLocation':
        """
        Sets the best bearing
        :param bearing: Bearing to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(bearing, [int, float])
        self._proto.bearing = bearing
        return self

    def get_vertical_accuracy(self) -> float:
        """
        :return: The best vertical accuracy
        """
        return self._proto.vertical_accuracy

    def set_vertical_accuracy(self, vertical_accuracy: float) -> 'BestLocation':
        """
        Sets the best vertical accuracy.
        :param vertical_accuracy: Accuracy to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(vertical_accuracy, [int, float])
        self._proto.vertical_accuracy = vertical_accuracy
        return self

    def get_horizontal_accuracy(self) -> float:
        """
        :return: The best horizontal accuracy
        """
        return self._proto.horizontal_accuracy

    def set_horizontal_accuracy(self, horizontal_accuracy: float) -> 'BestLocation':
        """
        Sets the best horizontal accuracy
        :param horizontal_accuracy: Accuracy to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(horizontal_accuracy, [int, float])
        self._proto.horizontal_accuracy = horizontal_accuracy
        return self

    def get_speed_accuracy(self) -> float:
        """
        :return: The best speed accuracy
        """
        return self._proto.speed_accuracy

    def set_speed_accuracy(self, speed_accuracy: float) -> 'BestLocation':
        """
        Sets the best speed accuracy
        :param speed_accuracy: Accuracy to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(speed_accuracy, [int, float])
        self._proto.speed_accuracy = speed_accuracy
        return self

    def get_bearing_accuracy(self) -> float:
        """
        :return: The best bearing accuracy
        """
        return self._proto.bearing_accuracy

    def set_bearing_accuracy(self, bearing_accuracy: float) -> 'BestLocation':
        """
        Sets the best bearing accuracy
        :param bearing_accuracy: Accuracy to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(bearing_accuracy, [int, float])
        self._proto.bearing_accuracy = bearing_accuracy
        return self

    def get_score(self) -> float:
        """
        :return: The best location score
        """
        return self._proto.score

    def set_score(self, score: float) -> 'BestLocation':
        """
        Sets the best location score
        :param score: Score to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(score, [int, float])
        self._proto.score = score
        return self

    def get_method(self) -> LocationScoreMethod:
        """
        :return: The best location score method
        """
        # noinspection Mypy
        return LocationScoreMethod.from_proto(self._proto.method)

    def set_method(self, method: LocationScoreMethod) -> 'BestLocation':
        """
        Sets the best location score method.
        :param method: Method to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(method, [LocationScoreMethod])
        # noinspection Mypy
        self._proto.method = method.into_proto()
        return self

    def get_location_provider(self) -> LocationProvider:
        """
        :return: The location provide associated with the best location.
        """
        # noinspection Mypy
        return LocationProvider.from_proto(self._proto.location_provider)

    def set_location_provider(self, location_provider: LocationProvider) -> 'BestLocation':
        """
        Sets the location provider associated with the best location.
        :param location_provider: Location provider to set.
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(location_provider, [LocationProvider])
        # noinspection Mypy
        self._proto.location_provider = location_provider.into_proto()
        return self


class Location(redvox.api1000.common.generic.ProtoBase[RedvoxPacketM.Sensors.Location]):
    """
    Encapsulates location data and metadata
    """
    def __init__(self, proto: RedvoxPacketM.Sensors.Location):
        super().__init__(proto)
        self._timestamps: common.TimingPayload = common.TimingPayload(proto.timestamps)
        self._timestamps_gps: common.TimingPayload = common.TimingPayload(proto.timestamps_gps)
        self._latitude_samples: common.SamplePayload = common.SamplePayload(proto.latitude_samples)
        self._longitude_samples: common.SamplePayload = common.SamplePayload(proto.longitude_samples)
        self._altitude_samples: common.SamplePayload = common.SamplePayload(proto.altitude_samples)
        self._speed_samples: common.SamplePayload = common.SamplePayload(proto.speed_samples)
        self._bearing_samples: common.SamplePayload = common.SamplePayload(proto.bearing_samples)
        self._horizontal_accuracy_samples: common.SamplePayload = common.SamplePayload(
            proto.horizontal_accuracy_samples)
        self._vertical_accuracy_samples: common.SamplePayload = common.SamplePayload(proto.vertical_accuracy_samples)
        self._speed_accuracy_samples: common.SamplePayload = common.SamplePayload(proto.speed_accuracy_samples)
        self._bearing_accuracy_samples: common.SamplePayload = common.SamplePayload(proto.bearing_accuracy_samples)
        self._last_best_location: BestLocation = BestLocation(proto.last_best_location)
        self._overall_best_location: BestLocation = BestLocation(proto.overall_best_location)
        # noinspection Mypy
        self._location_providers: ProtoRepeatedMessage = ProtoRepeatedMessage(
            proto,
            proto.location_providers,
            "location_providers",
            LocationProvider.from_proto,
            LocationProvider.into_proto,
        )

    @staticmethod
    def new() -> 'Location':
        """
        :return: A new, empty Location instance
        """
        return Location(RedvoxPacketM.Sensors.Location())

    def get_sensor_description(self) -> str:
        """
        :return: The location sensor description
        """
        return self._proto.sensor_description

    def set_sensor_description(self, sensor_description: str) -> 'Location':
        """
        Sets the location sensor's description
        :param sensor_description: Description to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(sensor_description, [str])
        self._proto.sensor_description = sensor_description
        return self

    def get_timestamps(self) -> common.TimingPayload:
        """
        :return: The timing payload (machine timestamps) associated with this sensor's location samples
        """
        return self._timestamps

    def get_timestamps_gps(self) -> common.TimingPayload:
        """
        :return: The timing payload (gps timestamps) associated with this sensor's location samples
        """
        return self._timestamps_gps

    def get_latitude_samples(self) -> common.SamplePayload:
        """
        :return: The SamplePayload containing latitude samples and statistics
        """
        return self._latitude_samples

    def get_longitude_samples(self) -> common.SamplePayload:
        """
        :return: The SamplePayload containing longitude samples and statistics
        """
        return self._longitude_samples

    def get_altitude_samples(self) -> common.SamplePayload:
        """
        :return: The SamplePayload containing altitude samples and statistics
        """
        return self._altitude_samples

    def get_speed_samples(self) -> common.SamplePayload:
        """
        :return: The SamplePayload containing speed samples and statistics
        """
        return self._speed_samples

    def get_bearing_samples(self) -> common.SamplePayload:
        """
        :return: The SamplePayload containing bearing samples and statistics
        """
        return self._bearing_samples

    def get_horizontal_accuracy_samples(self) -> common.SamplePayload:
        """
        :return: The SamplePayload containing horizontal accuracy samples and statistics
        """
        return self._horizontal_accuracy_samples

    def get_vertical_accuracy_samples(self) -> common.SamplePayload:
        """
        :return: The SamplePayload containing vertical accuracy samples and statistics
        """
        return self._vertical_accuracy_samples

    def get_speed_accuracy_samples(self) -> common.SamplePayload:
        """
        :return: The SamplePayload containing speed accuracy samples and statistics
        """
        return self._speed_accuracy_samples

    def get_bearing_accuracy_samples(self) -> common.SamplePayload:
        """
        :return: The SamplePayload containing bearing accuracy samples and statistics
        """
        return self._bearing_accuracy_samples

    def get_last_best_location(self) -> BestLocation:
        """
        :return: This returns the most recent BestLocation
        """
        return self._last_best_location

    def get_overall_best_location(self) -> BestLocation:
        """
        :return: This returns the overall best location since the station last moved
        """
        return self._overall_best_location

    def get_location_permissions_granted(self) -> bool:
        """
        :return: Have location permissions been granted?
        """
        return self._proto.location_permissions_granted

    def set_location_permissions_granted(self, location_permissions_granted: bool) -> 'Location':
        """
        Sets if location permissions have been granted.
        :param location_permissions_granted: Permissions to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(location_permissions_granted, [bool])
        self._proto.location_permissions_granted = location_permissions_granted
        return self

    def get_location_services_requested(self) -> bool:
        """
        :return: Have location services been requested by the user?
        """
        return self._proto.location_services_requested

    def set_location_services_requested(self, location_services_requested: bool) -> 'Location':
        """
        Sets if location services have been requested.
        :param location_services_requested:
        :return:
        """
        redvox.api1000.common.typing.check_type(location_services_requested, [bool])
        self._proto.location_services_requested = location_services_requested
        return self

    def get_location_services_enabled(self) -> bool:
        """
        :return: If location services are enabled.
        """
        return self._proto.location_services_enabled

    def set_location_services_enabled(self, location_services_enabled: bool) -> 'Location':
        """
        Sets if location servers are enabled.
        :param location_services_enabled: Set enabled or disabled
        :return: Modified version of self
        """
        redvox.api1000.common.typing.check_type(location_services_enabled, [bool])
        self._proto.location_services_enabled = location_services_enabled
        return self

    def get_location_providers(self) -> ProtoRepeatedMessage:
        """
        :return: A list of location providers associated with each location sample
        """
        return self._location_providers


def validate_location(loc_sensor: Location) -> List[str]:
    """
    Validates the location
    :param loc_sensor: Location to validate
    :return: A list of validation errors
    """
    errors_list = common.validate_timing_payload(loc_sensor.get_timestamps())
    if loc_sensor.get_latitude_samples().get_unit() != common.Unit.DECIMAL_DEGREES:
        errors_list.append("Location sensor latitude units are not decimal degrees")
    errors_list.extend(common.validate_sample_payload(loc_sensor.get_latitude_samples(), "Latitude"))

    if loc_sensor.get_longitude_samples().get_unit() != common.Unit.DECIMAL_DEGREES:
        errors_list.append("Location sensor longitude units are not decimal degrees")
    errors_list.extend(common.validate_sample_payload(loc_sensor.get_longitude_samples(), "Longitude"))

    if loc_sensor.get_altitude_samples().get_unit() != common.Unit.METERS:
        errors_list.append("Location sensor altitude units are not meters")
    errors_list.extend(common.validate_sample_payload(loc_sensor.get_altitude_samples(), "Altitude"))

    # todo: any of the below can be turned off as needed
    if loc_sensor.get_speed_samples().get_unit() != common.Unit.METERS_PER_SECOND:
        errors_list.append("Location sensor speed units are not meters per second")
    errors_list.extend(common.validate_sample_payload(loc_sensor.get_speed_samples(), "Speed"))

    if loc_sensor.get_bearing_samples().get_unit() != common.Unit.DECIMAL_DEGREES:
        errors_list.append("Location sensor bearing units are not decimal degrees")
    errors_list.extend(common.validate_sample_payload(loc_sensor.get_bearing_samples(), "Bearing"))

    if loc_sensor.get_horizontal_accuracy_samples().get_unit() != common.Unit.METERS:
        errors_list.append("Location sensor horizontal accuracy units are not meters")
    errors_list.extend(common.validate_sample_payload(loc_sensor.get_horizontal_accuracy_samples(),
                                                      "Horizontal Accuracy"))

    if loc_sensor.get_vertical_accuracy_samples().get_unit() != common.Unit.METERS:
        errors_list.append("Location sensor vertical accuracy units are not meters")
    errors_list.extend(common.validate_sample_payload(loc_sensor.get_vertical_accuracy_samples(),
                                                      "Vertical Accuracy"))

    if loc_sensor.get_speed_accuracy_samples().get_unit() != common.Unit.METERS_PER_SECOND:
        errors_list.append("Location sensor speed accuracy units are not meters per second")
    errors_list.extend(common.validate_sample_payload(loc_sensor.get_speed_accuracy_samples(),
                                                      "Speed Accuracy"))

    if loc_sensor.get_bearing_accuracy_samples().get_unit() != common.Unit.DECIMAL_DEGREES:
        errors_list.append("Location sensor bearing accuracy units are not decimal degrees")
    errors_list.extend(common.validate_sample_payload(loc_sensor.get_bearing_accuracy_samples(),
                                                      "Bearing Accuracy"))

    return errors_list



from datetime import datetime
from glob import glob
import os.path
from pathlib import Path
from typing import List, Optional

from PySide2.QtCore import Slot
from PySide2.QtGui import QFont
from PySide2.QtWidgets import QAbstractItemView, QTreeWidgetItem, QTreeWidget, QWidget
from dataclasses import dataclass
from functools import total_ordering

from redvox.api1000.wrapped_redvox_packet.wrapped_packet import WrappedRedvoxPacketM


@total_ordering
@dataclass
class RdvxmFile:
    path: str
    station_id: str
    date: datetime

    @staticmethod
    def from_path(path: str) -> 'RdvxmFile':
        filename: str = Path(path).name
        base: str = filename.split(".")[0]
        split_base: List[str] = base.split("_")
        station_id: str = split_base[0]
        ts_us: int = int(split_base[1])
        date: datetime = datetime.utcfromtimestamp(ts_us / 1E6)

        return RdvxmFile(path, station_id, date)

    def filename(self) -> str:
        return Path(self.path).name

    def read(self) -> WrappedRedvoxPacketM:
        return WrappedRedvoxPacketM.from_compressed_path(self.path)

    def __eq__(self, other) -> bool:
        return self.date == other.date

    def __lt__(self, other: 'RdvxmFile') -> bool:
        return self.date < other.date


def find_rdvxm_files(base_dir: str, structured: bool = False) -> List[RdvxmFile]:
    if structured:
        raise RuntimeError("Structured layout is not yet implemented")

    paths: List[str] = glob(os.path.join(base_dir, "*.rdvxm"))
    return sorted(list(map(RdvxmFile.from_path, paths)))


class FileExplorer(QTreeWidget):
    def __init__(self, base_dir: str, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.setColumnCount(3)
        self.setHeaderLabels([
            "Path",
            "Station",
            "Date"
        ])
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        # self.setFont(QFont("Courier"))

        self.rdvxm_files: List[RdvxmFile] = find_rdvxm_files(base_dir)

        for i, rdvxm_file in enumerate(self.rdvxm_files):
            self.insertTopLevelItem(i, QTreeWidgetItem([
                rdvxm_file.filename(),
                rdvxm_file.station_id,
                str(rdvxm_file.date)
            ]))

        self.resizeColumnToContents(0)
        self.resizeColumnToContents(1)

        self.itemSelectionChanged.connect(self.file_selected)

    @Slot()
    def file_selected(self):
        idx: int = self.selectedIndexes()[0].row()
        packet = self.rdvxm_files[idx].read()
        self.parentWidget().packet_explorer.update_from_packet(packet)
        # self.parentWidget().parentWidget().details_column.values_widget.update_with_values(list(packet.get_sensors().get_audio().get_samples().get_values()))



"""
This module provides decorators used throughout the API M SDK codebase.
"""
from enum import Enum

from google.protobuf.internal.enum_type_wrapper import EnumTypeWrapper


def wrap_enum(proto_type: EnumTypeWrapper):
    """
    Decorator that provides automatic conversion methods to convert between the protobuf enum and the SDK provided enum.
    Will add an into_proto(self) method and a from_proto(proto) dynamically to Python standard enum.Enum classes.
    :param proto_type: The protobuf type
    :return: A function for decorating Python enums.
    """
    def __wrap_enum(enum: Enum) -> Enum:
        """
        Wrapper function for decorating enums.
        :param enum: The Python enum to decorate.
        :return: The decorated enumeration with dynamic methods set
        """
        setattr(enum, "into_proto", lambda self: proto_type.Value(self.name))
        # noinspection Mypy, PyCallingNonCallable
        enum.from_proto = staticmethod(lambda proto: enum(proto))
        return enum

    return __wrap_enum






"""
This module provides functionality for working with single channel API M sensors.
"""

import redvox.api1000.common.common as common
import redvox.api1000.common.typing
import redvox.api1000.proto.redvox_api_m_pb2 as redvox_api_m_pb2

from typing import List, Optional

import redvox.api1000.common.generic


class Single(
        redvox.api1000.common.generic.ProtoBase[redvox_api_m_pb2.RedvoxPacketM.Sensors.Single]):
    """
    This class encapsulates single channel sensors data and metadata
    """
    def __init__(self, proto: redvox_api_m_pb2.RedvoxPacketM.Sensors.Single):
        super().__init__(proto)
        self._timestamps: common.TimingPayload = common.TimingPayload(proto.timestamps)
        self._samples: common.SamplePayload = common.SamplePayload(proto.samples)

    @staticmethod
    def new() -> 'Single':
        """
        :return: A new, empty Single sensor instance
        """
        return Single(redvox_api_m_pb2.RedvoxPacketM.Sensors.Single())

    def get_sensor_description(self) -> str:
        """
        :return: The sensor description
        """
        return self._proto.sensor_description

    def set_sensor_description(self, sensor_description: str) -> 'Single':
        """
        Sets the sensor description.
        :param sensor_description: Description to set.
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(sensor_description, [str])
        self._proto.sensor_description = sensor_description
        return self

    def get_timestamps(self) -> common.TimingPayload:
        """
        :return: The TimingPayload which contains timestamps associated with each sample
        """
        return self._timestamps

    def get_samples(self) -> common.SamplePayload:
        """
        :return: The sample payload
        """
        return self._samples


def validate_single(single_sensor: Single, payload_unit: Optional[common.Unit] = None) -> List[str]:
    """
    Validates the single channel sensor.
    :param single_sensor: Sensor to validate.
    :param payload_unit: Expected unit
    :return: A list of validated errors
    """
    errors_list = common.validate_timing_payload(single_sensor.get_timestamps())
    errors_list.extend(common.validate_sample_payload(single_sensor.get_samples(),
                                                      single_sensor.get_sensor_description(), payload_unit))
    return errors_list



"""
This module contains functions for reading and writing bulk API M data.
"""
import os
import glob
import redvox.api1000.common.lz4
import numpy as np
import redvox.api1000.wrapped_redvox_packet.wrapped_packet as api_m_wp
import redvox.common.date_time_utils as date_time_utils
from typing import List, Optional, Set, Tuple, Callable, TypeVar


REDVOX_API_M_FILE_EXT = "rdvxm"


def calc_evenly_sampled_timestamps(start: float, samples: int, rate_hz: float) -> np.array:
    """
    given a start time, calculates samples amount of evenly spaced timestamps at rate_hz
    :param start: float, start timestamp
    :param samples: int, number of samples
    :param rate_hz: float, sample rate in hz
    :return: np.array with evenly spaced timestamps starting at start
    """
    return np.array(start + date_time_utils.seconds_to_microseconds(np.arange(0, samples) / rate_hz))


# todo: stream inputs and read multiple buffers
class ReadWrappedPackets:
    """
    A searchable/sortable list of continuous wrapped API M redvox packets
    A set of API M packets are continuous if:
        * their redvox ids are equal
        * their uuids are equal
        * their app start machine timestamps are equal
        * their sample rate is constant
        * their sensors do not change
        * there is no considerable gap between the end of one packet and the start of the next
    properties:
        wrapped_packets: list of wrapped redvox API M packets
        redvox_id: string, redvox id of all the packets
        uuid: string, uuid of all the packets
        start_mach_timestamp: float, app_start_mach_timestamp of all the packets
        audio_sample_rate: float, sample rate of the audio sensor
    """
    def __init__(self, wrapped_packets: List[api_m_wp.WrappedRedvoxPacketM]):
        """
        initialize ReadWrappedPackets
        :param wrapped_packets: the packets to add; must be continuous
        """
        # set the values used to check continuity
        self.redvox_id: str = wrapped_packets[0].get_station_information().get_id()
        self.uuid: str = wrapped_packets[0].get_station_information().get_uuid()
        self.start_mach_timestamp: float = wrapped_packets[0].get_timing_information().get_app_start_mach_timestamp()
        self.audio_sample_rate: float = wrapped_packets[0].get_sensors().get_audio().get_sample_rate()
        # set the packets
        self.wrapped_packets: List[api_m_wp.WrappedRedvoxPacketM] = wrapped_packets
        if len(wrapped_packets) > 1:
            self.wrapped_packets = self.sort_packets()

    @staticmethod
    def _default_sort_packets(packet: api_m_wp.WrappedRedvoxPacketM):
        return packet.get_timing_information().get_packet_start_mach_timestamp()

    # this is for below to allow custom criteria to sort on
    T = TypeVar("T")

    def sort_packets(self, sort_func: Optional[Callable[[api_m_wp.WrappedRedvoxPacketM], T]] = None,
                     reverse: bool = False) -> List[api_m_wp.WrappedRedvoxPacketM]:
        """
        sort packets by custom user function, or by default, packet_start_timestamp
        :param sort_func: Optional function defining how to sort the packets, default None (uses packet start mach time)
        :param reverse: bool, if True, sort results in reverse, default False
        :return: the sorted list of packets
        """
        if sort_func is None:
            sort_func = self._default_sort_packets
        return sorted(self.wrapped_packets, key=sort_func, reverse=reverse)

    def add_packet(self, wrapped_packet: api_m_wp.WrappedRedvoxPacketM) -> bool:
        """
        Adds the wrapped packet to the list if it is from the same device
        :param wrapped_packet: packet to potentially add to list
        :return: True if successful, False otherwise
        """
        if wrapped_packet.get_station_information().get_id() == self.redvox_id \
                and wrapped_packet.get_station_information().get_uuid() == self.uuid \
                and wrapped_packet.get_timing_information().get_app_start_mach_timestamp() == self.start_mach_timestamp\
                and self.validate_sensors(wrapped_packet):
            self.wrapped_packets.append(wrapped_packet)
            self.wrapped_packets = self.sort_packets()
            return True
        return False

    def validate_sensors(self, wrapped_packet: api_m_wp.WrappedRedvoxPacketM) -> bool:
        """
        Checks for difference between sensors of the wrapped_packet and the ReadWrappedPacket
        :param wrapped_packet: packet to check for differences
        :return: True if no change in any sensors, False otherwise
        """
        # only 1 packet means sensors don't change
        if len(self.wrapped_packets) <= 1:
            return True
        prev_sensors = [wrapped_packet.get_sensors().has_audio(),
                        wrapped_packet.get_sensors().has_pressure(),
                        wrapped_packet.get_sensors().has_accelerometer(),
                        wrapped_packet.get_sensors().has_gyroscope(),
                        wrapped_packet.get_sensors().has_light(),
                        wrapped_packet.get_sensors().has_image(),
                        wrapped_packet.get_sensors().has_location(),
                        wrapped_packet.get_sensors().has_magnetometer(),
                        wrapped_packet.get_sensors().has_ambient_temperature(),
                        wrapped_packet.get_sensors().has_compress_audio(),
                        wrapped_packet.get_sensors().has_gravity(),
                        wrapped_packet.get_sensors().has_linear_acceleration(),
                        wrapped_packet.get_sensors().has_orientation(),
                        wrapped_packet.get_sensors().has_proximity(),
                        wrapped_packet.get_sensors().has_relative_humidity(),
                        wrapped_packet.get_sensors().has_rotation_vector()]
        for packet in self.wrapped_packets:
            next_sensors = [packet.get_sensors().has_audio(),
                            packet.get_sensors().has_pressure(),
                            packet.get_sensors().has_accelerometer(),
                            packet.get_sensors().has_gyroscope(),
                            packet.get_sensors().has_light(),
                            packet.get_sensors().has_image(),
                            packet.get_sensors().has_location(),
                            packet.get_sensors().has_magnetometer(),
                            packet.get_sensors().has_ambient_temperature(),
                            packet.get_sensors().has_compress_audio(),
                            packet.get_sensors().has_gravity(),
                            packet.get_sensors().has_linear_acceleration(),
                            packet.get_sensors().has_orientation(),
                            packet.get_sensors().has_proximity(),
                            packet.get_sensors().has_relative_humidity(),
                            packet.get_sensors().has_rotation_vector()]
            if prev_sensors != next_sensors:
                return False
        # if here, all sensors match
        return True

    def identify_gaps(self, allowed_timing_error_s: float, debug: bool = False) -> List[int]:
        """
        Identifies discontinuities in sensor data by checking if sensors drop in and out and by comparing timing info.
        :param allowed_timing_error_s: float, the amount of timing error in seconds.
        :param debug: bool, if True, output information as function is running, default False
        :return: A list of indices into the original list where gaps were found.
        """
        if len(self.wrapped_packets) <= 1:
            return []

        gaps = set()

        for i in range(1, len(self.wrapped_packets)):
            prev_packet = self.wrapped_packets[i - 1]
            next_packet = self.wrapped_packets[i]

            prev_timestamp = prev_packet.get_timing_information().get_packet_end_mach_timestamp()
            next_timestamp = next_packet.get_timing_information().get_packet_start_mach_timestamp()
            if debug:
                print(next_timestamp - prev_timestamp)
            # timestamps of audio data should be close together
            if date_time_utils.microseconds_to_seconds(next_timestamp - prev_timestamp) > allowed_timing_error_s:
                gaps.add(i)
                if debug:
                    print("time gap")

        return sorted(list(gaps))


class ReadResult:
    """
    Results from reading a directory containing API M redvox data
    properties:
        start_timestamp: optional float, start timestamp in microseconds of the data being read
        end_timestamp: optional float, end timestamp in microseconds of the data being read
        all_wrapped_packets: list of lists of wrapped API M redvox packets
    """
    def __init__(self, start_time: Optional[float] = None, end_time: Optional[float] = None,
                 wrapped_packets: Optional[List[ReadWrappedPackets]] = None):
        """
        initialize a ReadResult
        :param start_time: start time of the data being read, default None
        :param end_time: end time of the data being read, default None
        :param wrapped_packets: list of lists of wrapped API M packets containing all the data, default None
        """
        self.start_timestamp: Optional[float] = start_time
        self.end_timestamp: Optional[float] = end_time
        if wrapped_packets is None:
            self.all_wrapped_packets = []
        else:
            self.all_wrapped_packets = wrapped_packets

    def add_result(self, wrapped_packets: ReadWrappedPackets):
        """
        adds a ReadWrappedPackets object to the ReadResult
        :param wrapped_packets: a ReadWrappedPackets object
        """
        self.all_wrapped_packets.append(wrapped_packets)

    def add_packet(self, wrapped_packet: api_m_wp.WrappedRedvoxPacketM):
        """
        adds a single wrapped API M redvox packet to the ensemble
        :param wrapped_packet: a single wrapped API M redvox packet
        """
        for packets in self.all_wrapped_packets:
            if packets.add_packet(wrapped_packet):
                # packet got added, stop function
                return
        # went through for loop and nothing got added, so we assume it's a new id to add
        self.all_wrapped_packets.append(ReadWrappedPackets([wrapped_packet]))

    def add_list_of_packets(self, wrapped_packets: List[api_m_wp.WrappedRedvoxPacketM]):
        """
        adds a list of WrappedRedvoxPacketM files to the ReadResult
        :param wrapped_packets: the list of wrapped API M redvox packets to add
        """
        for packet in wrapped_packets:
            self.add_packet(packet)
        end_result = self.reorganize_packets()
        self.all_wrapped_packets = end_result.all_wrapped_packets

    def identify_gaps(self, timing_gap_s: float = 5.0) -> 'ReadResult':
        """
        checks all_wrapped_packets for any time gaps and splits them into continuous objects
        returns a copy of the calling object if no gaps detected
        :param timing_gap_s: amount of seconds allowed between packets to be considered a gap
        :return: an updated ReadResult object
        """
        updated_result = ReadResult(self.start_timestamp, self.end_timestamp)
        for packets in self.all_wrapped_packets:
            gaps = packets.identify_gaps(timing_gap_s)
            start = 0
            for index in gaps:
                split_val = ReadWrappedPackets(packets.wrapped_packets[start:index])
                updated_result.all_wrapped_packets.append(split_val)
                start = index
            # do this one last time to get the rest of the data that wasn't split yet
            split_val = ReadWrappedPackets(packets.wrapped_packets[start:])
            updated_result.all_wrapped_packets.append(split_val)
        return updated_result

    def reorganize_packets(self, timing_gap_s: float = 5.0) -> 'ReadResult':
        """
        takes all_wrapped_packets and recreates the ReadResult object
        :param timing_gap_s: amount of seconds allowed between packets to be considered a gap
        :return: an updated ReadResult object
        """
        updated_result = ReadResult(self.start_timestamp, self.end_timestamp)
        for packets in self.all_wrapped_packets:
            for packet in packets.wrapped_packets:
                updated_result.add_packet(packet)
        return updated_result.identify_gaps(timing_gap_s)

    def get_by_id(self, redvox_id: str) -> List[api_m_wp.WrappedRedvoxPacketM]:
        """
        gets all the packets with the id specified
        :param redvox_id: the redvox id to return values for
        :return: list of API M redvox packets with the redvox_id specified
        """
        results: List[api_m_wp.WrappedRedvoxPacketM] = []
        for packets in self.all_wrapped_packets:
            if packets.redvox_id == redvox_id:
                results.extend(packets.wrapped_packets)
        return results


class StreamResult:
    pass


def wrap(redvox_packet: api_m_wp.RedvoxPacketM) -> api_m_wp.WrappedRedvoxPacketM:
    """
    Wraps a protobuf packet in a WrappedRedocPacket.
    :param redvox_packet: Protobuf packet to wrap.
    :return: A WrappedRedvoxPacket.
    """
    return api_m_wp.WrappedRedvoxPacketM(redvox_packet)


def read_buffer(buf: bytes, is_compressed: bool = True) -> api_m_wp.RedvoxPacketM:
    """
    Deserializes a serialized protobuf RedvoxPacket buffer.
    :param buf: Buffer to deserialize.
    :param is_compressed: Whether or not the buffer is compressed or decompressed.
    :return: Deserialized protobuf redvox packet.
    """
    buffer = redvox.api1000.common.lz4.decompress(buf) if is_compressed else buf
    redvox_packet = api_m_wp.RedvoxPacketM()
    redvox_packet.ParseFromString(buffer)
    return redvox_packet


def read_file(file: str, is_compressed: bool = None) -> api_m_wp.RedvoxPacketM:
    """
    Deserializes a serialized protobuf RedvoxPacketM file.
    :param file: File to deserialize.
    :param is_compressed: Whether or not the file is compressed or decompressed.
    :return: Deserialized protobuf API M redvox packet.
    """
    file_ext = file.split(".")[-1]

    if is_compressed is None:
        _is_compressed = True if file_ext == REDVOX_API_M_FILE_EXT else False
    else:
        _is_compressed = is_compressed
    with open(file, "rb") as fin:
        return read_buffer(fin.read(), _is_compressed)


def read_rdvxm_file(path: str) -> api_m_wp.WrappedRedvoxPacketM:
    """
    Reads a .rdvxm file from the specified path and returns a WrappedRedvoxPacketM.
    :param path: The path of the file.
    :return: A WrappedRedvoxPacketM.
    """
    return wrap(read_file(path))


def read_rdvxm_buffer(buf: bytes) -> api_m_wp.WrappedRedvoxPacketM:
    """
    reads a .rdvxm file from the provided buffer and returns a WrappedRedvoxPacketM
    :param buf: buffer of bytes to read
    :return: A WrappedRedvoxPacketM
    """
    return wrap(read_buffer(buf))


def _extract_timestamp(path: str) -> int:
    """
    Extracts a timestamp in microseconds from a file path.
    :param path: Path to extract a timestamp from.
    :return: The timestamp in seconds from a path.
    """
    file = path.split(os.path.sep)[-1].split(".")[0]
    return int(date_time_utils.microseconds_to_seconds(int(file.split("_")[1])))


def _extract_redvox_id(path: str) -> str:
    """
    Extracts a redvox id from a file path.
    :param path: A path to extract the id from.
    :return: The redvox id.
    """
    file = path.split(os.path.sep)[-1]
    return file.split("_")[0]


def _is_int(int_as_str: str) -> bool:
    """
    Returns true if the given string can be parsed as an int.
    :param int_as_str: String to test.
    :return: True if it is an int, False otherwise.
    """
    try:
        int(int_as_str)
        return True
    except ValueError:
        return False


def _is_valid_redvox_api_m_filename(filename: str) -> bool:
    """
    Given a filename, determine if the filename is a valid API M redvox file name.
    :param filename: Filename to test.
    :return: True if it is valid, false otherwise.
    """
    return len(filename) == 33 and _is_int(filename[0:10]) \
        and filename[10:11] == "_" and _is_int(filename[11:27]) \
        and filename[27:len(filename)] == f".{REDVOX_API_M_FILE_EXT}"


def _is_path_in_set(path: str,
                    start_timestamp_utc_s: int,
                    end_timestamp_utc_s: int,
                    redvox_ids: Set[str] = None) -> bool:
    """
    Determines whether a given path is in a provided time range and set of redvox_ids.
    :param path: The path to check.
    :param start_timestamp_utc_s: Start of time range in seconds since epoch UTC.
    :param end_timestamp_utc_s: End of time range in seconds since epoch UTC.
    :param redvox_ids: Optional set of redvox ids.
    :return: True if path is in set false otherwise.
    """
    if redvox_ids is None:
        redvox_ids = set()
    filename = path.split(os.sep)[-1]

    if not _is_valid_redvox_api_m_filename(filename):
        return False

    timestamp = int(date_time_utils.microseconds_to_seconds(float(filename[11:27])))

    if not start_timestamp_utc_s <= timestamp <= end_timestamp_utc_s:
        return False

    if len(redvox_ids) > 0:
        if len(redvox_ids) > 0:
            redvox_id = filename[0:10]
            if redvox_id not in redvox_ids:
                return False

    return True


def _get_time_range_paths(paths: List[str],
                          redvox_ids: Optional[Set[str]]) -> Tuple[int, int]:
    """
    Given a set of paths, get the start and end timestamps.
    :param paths: Paths to get time range from.
    :param redvox_ids: Redvox ids to filter on.
    :return: A tuple containing the start and end timestamps of the data range.
    """
    if redvox_ids is not None:
        paths = list(filter(lambda path: _extract_redvox_id(path) in redvox_ids, paths))

    if len(paths) == 0:
        return -1, -1

    if len(paths) == 1:
        timestamp = _extract_timestamp(paths[0])
        return timestamp, timestamp

    timestamps = sorted(list(map(_extract_timestamp, paths)))
    return timestamps[0], timestamps[-1]


def _get_paths_time_range(directory: str,
                          redvox_ids: Optional[Set[str]],
                          recursive: bool) -> Tuple[int, int]:
    """
    Gets the start and end time range of a given set of API M data.
    :param directory: The base directory of the data.
    :param redvox_ids: Redvox ids to filter on.
    :param recursive: Whether or not to perform a recursive search.
    :return: A tuple containing the start and end timestamps of the given data set.
    """
    if recursive:
        all_paths = glob.glob(os.path.join(directory, "**", f"*.{REDVOX_API_M_FILE_EXT}"),
                              recursive=recursive)
    else:
        all_paths = glob.glob(os.path.join(directory, f"*.{REDVOX_API_M_FILE_EXT}"),
                              recursive=recursive)

    return _get_time_range_paths(all_paths, redvox_ids)


def _get_structured_paths(directory: str,
                          start_timestamp_utc_s: int,
                          end_timestamp_utc_s: int,
                          redvox_ids: Set[str] = None) -> List[str]:
    """
    Given a base directory (which should end with api1000), find the paths of all structured API M redvox files.
    :param directory: The base directory path (which should end with api1000)
    :param start_timestamp_utc_s: Start timestamp as seconds since the epoch UTC.
    :param end_timestamp_utc_s: End timestamp as seconds since the epoch UTC.
    :param redvox_ids: An optional set of redvox_ids to filter against.
    :return: A list of paths in a structured layout of filtered API M redvox files.
    """
    if redvox_ids is None:
        redvox_ids = set()
    paths = []
    for (year, month, day, hour) in date_time_utils.DateIteratorAPIM(start_timestamp_utc_s, end_timestamp_utc_s):
        all_paths = glob.glob(os.path.join(directory, year, month, day, hour, f"*.{REDVOX_API_M_FILE_EXT}"))
        valid_paths = list(
            filter(lambda path: _is_path_in_set(path, start_timestamp_utc_s, end_timestamp_utc_s, redvox_ids),
                   all_paths))
        paths.extend(valid_paths)
    return paths


def read_structured(directory: str,
                    start_timestamp_utc_s: Optional[int] = None,
                    end_timestamp_utc_s: Optional[int] = None,
                    redvox_ids: Optional[List[str]] = None,
                    structured_layout: bool = True) -> ReadResult:
    """
    read API M data from a directory that contains a specific structure (directory/YYYY/MM/DD/HH)
    Timestamps in directory are in microseconds since epoch UTC
    :param directory: the root directory that contains the data
    :param start_timestamp_utc_s: starting timestamp in seconds since epoch UTC.  if None, get all data, default None
    :param end_timestamp_utc_s: ending timestamp in seconds since epoch UTC.  if None, get all data, default None
    :param redvox_ids: specific redvox ids to get data for.  if None, get all ids, default None
    :param structured_layout: specifies if directory has specific structure, default True
    :return: a ReadResult object that stores the packets with metadata
    """
    # Remove trailing directory separators
    if redvox_ids is None:
        redvox_ids = []
    while directory.endswith("/") or directory.endswith("\\"):
        directory = directory[:-1]

    if start_timestamp_utc_s is None or end_timestamp_utc_s is None:
        ids = None if len(redvox_ids) == 0 else redvox_ids
        start_adjusted, end_adjusted = _get_paths_time_range(directory, ids, structured_layout)

        if start_timestamp_utc_s is None:
            start_timestamp_utc_s = start_adjusted

        if end_timestamp_utc_s is None:
            end_timestamp_utc_s = end_adjusted

    if structured_layout:
        paths = _get_structured_paths(directory,
                                      start_timestamp_utc_s,
                                      end_timestamp_utc_s,
                                      set(redvox_ids))
    else:
        all_paths = glob.glob(os.path.join(directory, f"*.{REDVOX_API_M_FILE_EXT}"))
        paths = list(
            filter(lambda pth: _is_path_in_set(pth, start_timestamp_utc_s, end_timestamp_utc_s, set(redvox_ids)),
                   all_paths))

    read_result = ReadResult(start_timestamp_utc_s, end_timestamp_utc_s)

    for path in paths:
        read_result.add_packet(read_rdvxm_file(path))

    return read_result


def read_dir(directory_path: str) -> ReadResult:
    """
    read API M data from a single directory of unsorted API M redvox data
    :param directory_path: the directory path that contains all the data
    :return: the data as a ReadResult
    """
    # Make sure the directory ends with a trailing slash "/"
    if directory_path[-1] != "/":
        directory_path = directory_path + "/"

    read_result = ReadResult()

    file_paths = sorted(glob.glob(directory_path + f"*.{REDVOX_API_M_FILE_EXT}"))

    for path in file_paths:
        read_result.add_packet(read_rdvxm_file(path))

    return read_result


def stream_structured():
    pass


def stream_dir():
    pass


def read_buffers():
    pass



"""
Modules for extracting time synchronization statistics for API 900 data.
Also includes functions for correcting time arrays.
ALL timestamps in microseconds unless otherwise stated
"""

# noinspection Mypy
import numpy as np
import pandas as pd

from typing import List, Optional
from redvox.common import stats_helper as sh, tri_message_stats as tms, date_time_utils as dt
from redvox.common import sensor_data as sd


class TimeSyncData:
    """
    Stores latencies, offsets, and other timesync related information about a single station
    ALL timestamps in microseconds unless otherwise stated
    properties:
        station_id: str, id of station, default empty string
        station_start_timestamp: float, timestamp of when the station was started, default np.nan
        sample_rate_hz: float, sample rate of audio sensor in Hz, default np.nan
        packet_start_time: float, timestamp of when data started recording, default np.nan
        packet_end_time: float, timestamp of when data stopped recording, default np.nan
        packet_duration: float, length of packet in microseconds, default 0.0
        server_acquisition_time: float, timestamp of when packet arrived at server, default np.nan
        time_sync_exchanges_df: dataframe, timestamps that form the time synch exchanges, default empty dataframe
        latencies: np.ndarray, calculated latencies of the exchanges, default empty np.ndarray
        best_latency_index: int, index in latencies array that contains the best latency, default np.nan
        best_latency: float, best latency of the data, default np.nan
        mean_latency: float, mean latency, default np.nan
        latency_std: float, standard deviation of latencies, default np.nan
        offsets: np.ndarray, calculated offsets of the exchanges, default np.ndarray
        best_offset: float, best offset of the data, default np.nan
        mean_offset: float, mean offset, default np.nan
        offset_std: float, standard deviation of offsets, default np.nan
        best_tri_msg_index: int, index of the best tri-message, default np.nan
        acquire_travel_time: float, calculated time it took packet to reach server, default np.nan
    """

    def __init__(self, data_pack: Optional[sd.DataPacket] = None,
                 station_metadata: Optional[sd.StationMetadata] = None):
        """
        Initialize properties
        :param data_pack: data packet metadata
        :param station_metadata: station metadata
        """
        self.best_latency: float = np.nan
        self.best_offset: float = 0
        if station_metadata is not None and data_pack is not None:
            self.get_timesync_data(data_pack, station_metadata)
        else:
            if station_metadata is None:
                self.station_id: str = ""
                self.station_start_timestamp: float = np.nan
            if data_pack is None:
                self.sample_rate_hz: float = np.nan
                self.packet_num_audio_samples: int = 0
                self.server_acquisition_time: float = np.nan
                self.packet_start_time: float = np.nan
                self.packet_end_time: float = np.nan
                self.packet_duration: float = 0.0
                self.time_sync_exchanges_df = pd.DataFrame([], columns=["a1", "a2", "a3", "b1", "b2", "b3"])
                self.best_tri_msg_index: int = np.nan
                self.latencies: np.ndarray = np.ndarray((0, 0))
                self.best_latency_index: int = np.nan
                self.mean_latency: float = np.nan
                self.latency_std: float = np.nan
                self.offsets: np.ndarray = np.ndarray((0, 0))
                self.mean_offset: float = 0.0
                self.offset_std: float = 0.0
                self.acquire_travel_time: float = np.nan
                # self.num_packets: int = 0
                # self.bad_packets: List[int] = []

    def get_timesync_data(self, data_pack: Optional[sd.DataPacket] = None,
                          station_metadata: Optional[sd.StationMetadata] = None):
        """
        extracts the time sync data from the data_pack object
        :param data_pack: data packet metadata
        :param station_metadata: station metadata
        """
        self.station_id = station_metadata.station_id
        self.sample_rate_hz = station_metadata.timing_data.audio_sample_rate_hz
        self.packet_num_audio_samples = data_pack.packet_num_audio_samples
        self.station_start_timestamp = station_metadata.timing_data.station_start_timestamp
        self.server_acquisition_time = data_pack.server_timestamp
        self.packet_start_time = data_pack.data_start_timestamp
        self.packet_end_time = data_pack.data_end_timestamp
        if data_pack.timesync is not None:
            self.time_sync_exchanges_df = pd.DataFrame(
                tms.transmit_receive_timestamps_microsec(data_pack.timesync),
                index=["a1", "a2", "a3", "b1", "b2", "b3"]).T
        else:
            self.time_sync_exchanges_df = pd.DataFrame([], columns=["a1", "a2", "a3", "b1", "b2", "b3"])
        # stations may contain the best latency and offset data already
        if data_pack.packet_best_latency:
            self.best_latency = data_pack.packet_best_latency
        if data_pack.packet_best_offset:
            self.best_offset = data_pack.packet_best_offset
        self._compute_tri_message_stats()
        # set the packet duration
        self.packet_duration = self.packet_end_time - self.packet_start_time
        # calculate travel time between corrected end of packet timestamp and server timestamp
        self.acquire_travel_time = self.server_acquisition_time - (self.packet_end_time + self.best_offset)

    def _compute_tri_message_stats(self):
        """
        Compute the tri-message stats from the data
        """
        if self.num_tri_messages() > 0:
            # compute tri message data from time sync exchanges
            tse = tms.TriMessageStats(self.station_id,
                                      self.time_sync_exchanges_df["a1"], self.time_sync_exchanges_df["a2"],
                                      self.time_sync_exchanges_df["a3"],
                                      self.time_sync_exchanges_df["b1"], self.time_sync_exchanges_df["b2"],
                                      self.time_sync_exchanges_df["b3"])
            # Compute the statistics for latency and offset
            self.mean_latency = np.mean([*tse.latency1, *tse.latency3])
            self.latency_std = np.std([*tse.latency1, *tse.latency3])
            self.mean_offset = np.mean([*tse.offset1, *tse.offset3])
            self.offset_std = np.std([*tse.offset1, *tse.offset3])
            self.latencies = np.array((tse.latency1, tse.latency3))
            self.offsets = np.array((tse.offset1, tse.offset3))
            self.best_latency_index = tse.best_latency_index
            self.best_tri_msg_index = tse.best_latency_index
            # if best_latency is None, set to best computed latency
            if np.isnan(self.best_latency):
                self.best_latency = tse.best_latency
                self.best_offset = tse.best_offset
            # if best_offset is still default value, use the best computed offset
            elif self.best_offset == 0:
                self.best_offset = tse.best_offset
        else:
            # If here, there are no exchanges to read.  writing default or empty values to the correct properties
            self.best_tri_msg_index = np.nan
            self.best_latency_index = np.nan
            self.best_latency = np.nan
            self.mean_latency = np.nan
            self.latency_std = np.nan
            self.best_offset = 0
            self.mean_offset = 0
            self.offset_std = 0

    def num_tri_messages(self) -> int:
        """
        return the number of tri-message exchanges
        :return: number of tri-message exchanges
        """
        return self.time_sync_exchanges_df.shape[0]


class TimeSyncAnalysis:
    """
    Used for multiple TimeSyncData objects from a station
    properties:
        station_id: string, the station_id of the station being analyzed, default empty string
        best_latency_index: int, the index of the TimeSyncData object with the best latency, default np.nan
        latency_stats: StatsContainer, the statistics of the latencies
        offset_stats: StatsContainer, the statistics of the offsets
        sample_rate_hz: float, the audio sample rate in hz of the station, default np.nan
        timesync_data: list of TimeSyncData, the TimeSyncData to analyze, default empty list
        station_start_timestamp: float, the timestamp of when the station became active, default np.nan
    """
    def __init__(self, station: Optional[sd.Station] = None):
        """
        Initialize properties
        :param station: the station to perform analysis on
        """
        self.station_id: str = ""
        self.best_latency_index: int = np.nan
        self.latency_stats = sh.StatsContainer("latency")
        self.offset_stats = sh.StatsContainer("offset")
        self.sample_rate_hz: float = np.nan
        self.timesync_data: List[TimeSyncData] = []
        self.station_start_timestamp: float = np.nan
        if station is not None:
            self.station_id = station.station_metadata.station_id
            self.station_start_timestamp = station.station_metadata.timing_data.station_start_timestamp
            self.sample_rate_hz = station.station_metadata.timing_data.audio_sample_rate_hz
            self.timesync_data = get_time_sync_data_from_station(station)
            self.evaluate_and_validate_data()

    def evaluate_and_validate_data(self):
        """
        check the data for errors and update the analysis statistics
        """
        self._calc_timesync_stats()
        self.evaluate_latencies()
        if self.validate_start_timestamp():
            self.station_start_timestamp = self.timesync_data[0].station_start_timestamp
        if self.validate_sample_rate():
            self.sample_rate_hz = self.timesync_data[0].sample_rate_hz

    def _calc_timesync_stats(self):
        """
        calculates the mean and std deviation for latencies and offsets
        """
        if len(self.timesync_data) < 1:
            raise ValueError("Nothing to calculate stats; length of timesync data is less than 1")
        # reset the stats containers
        self.latency_stats = sh.StatsContainer("latency")
        self.offset_stats = sh.StatsContainer("offset")
        for index in range(len(self.timesync_data)):
            # add the stats of the latency
            self.latency_stats.add(self.timesync_data[index].mean_latency, self.timesync_data[index].latency_std,
                                   self.timesync_data[index].num_tri_messages() * 2)
            # add the stats of the offset
            self.offset_stats.add(self.timesync_data[index].mean_offset, self.timesync_data[index].offset_std,
                                  self.timesync_data[index].num_tri_messages() * 2)

    def add_timesync_data(self, timesync_data: TimeSyncData):
        """
        adds a TimeSyncData object to the analysis
        :param timesync_data: TimeSyncData to add
        """
        self.timesync_data.append(timesync_data)
        self.evaluate_and_validate_data()

    def get_num_packets(self) -> int:
        """
        return the number of packets in the TimeSyncAnalysis
        :return: number of packets analyzed
        """
        return len(self.timesync_data)

    def get_best_latency(self) -> float:
        """
        return the best latency
        :return: the best latency
        """
        if np.isnan(self.best_latency_index):
            return np.nan
        return self.timesync_data[self.best_latency_index].best_latency

    def get_latencies(self) -> np.array:
        """
        return all the latencies
        :return: np.array containing all the latencies
        """
        latencies = []
        for ts_data in self.timesync_data:
            latencies.append(ts_data.best_latency)
        return np.array(latencies)

    def get_mean_latency(self) -> float:
        """
        return the mean of the latencies
        :return: the mean of the latencies, or np.nan if it doesn't exist
        """
        return self.latency_stats.mean_of_means()

    def get_latency_stdev(self) -> float:
        """
        return the standard deviation of the latencies
        :return: the standard deviation of the latencies, or np.nan if it doesn't exist
        """
        return self.latency_stats.total_std_dev()

    def get_best_offset(self) -> float:
        """
        returns the offset associated with the best latency
        :return: offset associated with the best latency
        """
        if np.isnan(self.best_latency_index):
            return np.nan
        return self.timesync_data[self.best_latency_index].best_offset

    def get_offsets(self) -> np.array:
        """
        return all the offsets
        :return: np.array containing all the offsets
        """
        offsets = []
        for ts_data in self.timesync_data:
            offsets.append(ts_data.best_offset)
        return np.array(offsets)

    def get_mean_offset(self) -> float:
        """
        return the mean of the offsets
        :return: the mean of the offsets, or np.nan if it doesn't exist
        """
        return self.offset_stats.mean_of_means()

    def get_offset_stdev(self) -> float:
        """
        return the standard deviation of the offsets
        :return: the standard deviation of the offsets, or np.nan if it doesn't exist
        """
        return self.offset_stats.total_std_dev()

    def get_best_packet_latency_index(self) -> int:
        """
        returns the index of the best latency in the packet with the best latency
        :return: the best latency's index in the packet with the best latency
        """
        if np.isnan(self.best_latency_index):
            return np.nan
        return self.timesync_data[self.best_latency_index].best_latency_index

    def get_best_start_time(self) -> float:
        """
        return the start timestamp associated with the best latency
        :return: start timestamp associated with the best latency
        """
        if np.isnan(self.best_latency_index):
            return np.nan
        return self.timesync_data[self.best_latency_index].packet_start_time

    def get_start_times(self) -> np.array:
        """
        returns the start timestamps of all the packets
        :return: list of the start timestamps of each packet
        """
        start_times = []
        for ts_data in self.timesync_data:
            start_times.append(ts_data.packet_start_time)
        return np.array(start_times)

    def get_bad_packets(self) -> List[int]:
        """
        returns all the packets that have invalid data
        :return: list of all packets that contains invalid data
        """
        bad_packets = []
        for idx in range(self.get_num_packets()):  # mark bad indices (they have a 0 or less value)
            if self.get_latencies()[idx] <= 0 or np.isnan(self.get_latencies()[idx]):
                bad_packets.append(idx)
        return bad_packets

    def evaluate_latencies(self):
        """
        finds the best latency
        outputs warnings if a change in timestamps is detected
        """
        if self.get_num_packets() < 1:
            raise ValueError("Nothing to evaluate; length of timesync data is less than 1")
        self.best_latency_index = 0
        # assume the first element has the best timesync values for now, then compare with the others
        for index in range(1, self.get_num_packets()):
            # find the best latency; in this case, the minimum
            # if new value is not None and if the current best is None or new value is better than current best, update
            if not np.isnan(self.timesync_data[index].best_latency) \
                    and (np.isnan(self.get_best_latency())
                         or self.timesync_data[index].best_latency < self.get_best_latency()):
                self.best_latency_index = index

    def validate_start_timestamp(self, debug: bool = False) -> bool:
        """
        confirms if station_start_timestamp differs in any of the timesync_data
        outputs warnings if a change in timestamps is detected
        :param debug: if True, output warning message, default False
        :return: True if no change
        """
        if self.get_num_packets() < 1:
            if debug:
                print("Less than 1 timesync data object to evaluate start timestamps with")
        for index in range(0, self.get_num_packets()):
            # compare station start timestamps; notify when they are different
            if self.timesync_data[index].station_start_timestamp != self.station_start_timestamp:
                if debug:
                    print(f"Warning!  Change in station start timestamp detected!  "
                          f"Expected: {self.station_start_timestamp}, read: "
                          f"{self.timesync_data[index].station_start_timestamp}")
                return False
        # if here, all the sample timestamps are the same
        return True

    def validate_sample_rate(self, debug: bool = False) -> bool:
        """
        confirms if sample rate is the same across all timesync_data
        outputs warning if a change in sample rate is detected
        :param debug: if True, output warning message, default False
        :return: True if no change
        """
        if self.get_num_packets() < 1:
            if debug:
                print("Less than 1 timesync data object to evaluate sample rate with")
        for index in range(0, self.get_num_packets()):
            # compare station start timestamps; notify when they are different
            if np.isnan(self.timesync_data[index].sample_rate_hz) \
                    or self.timesync_data[index].sample_rate_hz != self.sample_rate_hz:
                if debug:
                    print(f"Warning!  Change in station sample rate detected!  "
                          f"Expected: {self.sample_rate_hz}, read: {self.timesync_data[index].sample_rate_hz}")
                return False
        # if here, all the sample rates are the same
        return True

    def validate_time_gaps(self, gap_duration_s: float, debug: bool = False) -> bool:
        """
        confirms there are no data gaps between packets
        outputs warning if a gap is detected
        :param gap_duration_s: length of time in seconds to be detected as a gap
        :param debug: if True, output warning message, default False
        :return: True if no gap
        """
        if self.get_num_packets() < 2:
            if debug:
                print("Less than 2 timesync data objects to evaluate gaps with")
        for index in range(1, self.get_num_packets()):
            # compare last packet's end timestamp with current start timestamp
            if dt.microseconds_to_seconds(self.timesync_data[index].packet_start_time
                                          - self.timesync_data[index - 1].packet_end_time) > gap_duration_s:
                if debug:
                    print(f"Warning!  Gap detected at packet number: {index}")
                return False
        # if here, no gaps
        return True


def get_time_sync_data_from_station(station: sd.Station) -> List[TimeSyncData]:
    """
    Returns the TimeSyncData associated with a station
    :param station: the station to get data from
    :return: a list of all TimeSyncData objects from the station
    """
    timesync_list = []
    for packet in station.packet_data:
        timesync_list.append(TimeSyncData(packet, station.station_metadata))
    return timesync_list


def validate_sensors(tsa_data: TimeSyncAnalysis) -> bool:
    """
    Examine all sample rates and mach time zeros to ensure that sensor settings do not change
    :param tsa_data: the TimeSyncAnalysis data to validate
    :return: True if sensor settings do not change
    """
    # check that we have packets to read
    if tsa_data.get_num_packets() < 1:
        print("ERROR: no data to validate.")
        return False
    elif tsa_data.get_num_packets() > 1:
        # if we have more than one packet, we need to validate the data
        return tsa_data.validate_sample_rate() and tsa_data.validate_start_timestamp()
    # we get here if all packets have the same sample rate and mach time zero
    return True


def update_evenly_sampled_time_array(ts_analysis: TimeSyncAnalysis, num_samples: float = None,
                                     time_start_array_s: np.array = None) -> np.ndarray:
    """
    Correct evenly sampled times using updated time_start_array values as the focal point.
    Expects tsd to have the same number of packets as elements in time_start_array.
    Expects there are no gaps in the data or changes in station sample rate or start time.
    Throws an exception if the number of packets in tsa does not match the length of time_start_array
    :param ts_analysis: TimeSyncAnalysis object that contains the information needed to update the time array
    :param num_samples: number of samples in one file; optional, uses number based on sample rate if not given
    :param time_start_array_s: the array of timestamps to correct in seconds; optional, uses the start times in the
                             TimeSyncAnalysis object if not given
    :return: Revised time array in epoch seconds
    """
    if not validate_sensors(ts_analysis):
        raise AttributeError("ERROR: Change in Station Start Time or Sample Rate detected!")
    if time_start_array_s is None:
        # replace the time_start_array with values from tsd; convert tsd times to seconds
        time_start_array_s = np.array([])
        for tsd in ts_analysis.timesync_data:
            time_start_array_s = np.append(time_start_array_s, tsd.packet_start_time / dt.MICROSECONDS_IN_SECOND)
    num_files = len(ts_analysis.timesync_data)
    # the TimeSyncData must have the same number of packets as the number of elements in time_start_array
    if num_files != len(time_start_array_s):
        # alert the user, then quit
        raise Exception("ERROR: Attempted to update a time array that doesn't contain "
                        "the same number of elements as the TimeSyncAnalysis!")

    # use the number of audio samples in the first data packet
    if num_samples is None:
        num_samples = ts_analysis.timesync_data[0].packet_num_audio_samples
    t_dt = 1.0 / ts_analysis.sample_rate_hz

    # Use TimeSyncData object to find best start index.
    # Samples before will be the number of decoders before a0 times the number of samples in a file.
    # Samples after will be the number of decoders after a0 times the number of samples in a file minus 1;
    # the minus one represents the best a0.
    decoder_idx = ts_analysis.best_latency_index
    samples_before = int(decoder_idx * num_samples)
    samples_after = round((num_files - decoder_idx) * num_samples) - 1
    best_start_sec = time_start_array_s[decoder_idx]

    # build the time arrays separately in epoch seconds, then join into one
    # add 1 to include the actual a0 sample, then add 1 again to skip the a0 sample; this avoids repetition
    timesec_before = np.vectorize(lambda t: best_start_sec - t * t_dt)(list(range(int(samples_before + 1))))
    timesec_before = timesec_before[::-1]  # reverse 'before' times so they increase from earliest start time
    timesec_after = np.vectorize(lambda t: best_start_sec + t * t_dt)(list(range(1, int(samples_after + 1))))
    timesec_rev = np.concatenate([timesec_before, timesec_after])

    return update_time_array_from_analysis(ts_analysis, timesec_rev)


def update_time_array(ts_data: TimeSyncData, time_array_s: np.array) -> np.ndarray:
    """
    Correct timestamps in time_array using information from TimeSyncData
    :param ts_data: TimeSyncData object that contains the information needed to update the time array
    :param time_array_s: the list of timestamps to correct in seconds
    :return: Revised time array in epoch seconds
    """
    return time_array_s + (ts_data.best_offset / dt.MICROSECONDS_IN_SECOND)


def update_time_array_from_analysis(ts_analysis: TimeSyncAnalysis, time_array_s: np.array) -> np.ndarray:
    """
    Correct timestamps in time_array using information from TimeSyncAnalysis
    :param ts_analysis: TimeSyncAnalysis object that contains the information needed to update the time array
    :param time_array_s: the list of timestamps to correct in seconds
    :return: Revised time array in epoch seconds
    """
    return time_array_s + (ts_analysis.get_best_offset() / dt.MICROSECONDS_IN_SECOND)



"""
This module contains functions for performing RedVox data conversions and displaying the contents of rdvxz files.
"""

import logging
import os.path
from typing import List, Optional

import redvox.api900.reader as reader
import redvox.api900.reader_utils as reader_utils
import redvox.common.api_conversions as api_conversions

from redvox.api1000.wrapped_redvox_packet.wrapped_packet import WrappedRedvoxPacketM

# pylint: disable=C0103
log = logging.getLogger(__name__)


def validate_rdvxm(paths: List[str]) -> bool:
    """
    Validates the correctness of rdvxm files.
    :param paths: Paths to the files.
    :return: True if all valid, False otherwise
    """
    for path in paths:
        wrapped_packet: WrappedRedvoxPacketM = WrappedRedvoxPacketM.from_compressed_path(path)
        validation_results: List[str] = wrapped_packet.validate()

        if len(validation_results) > 0:
            print(f"{len(validation_results)} validation issues found for file at path {path}")
            for i, validation_result in enumerate(validation_results):
                print(f"{i} {validation_result}")

    return True


def rdvxz_to_json(paths: List[str], out_dir: Optional[str] = None) -> bool:
    """
    Converts .rdvxz files to .json files.
    :param paths: Paths of .rdvxz files to convert.
    :param out_dir: An optional output directory (will use input directory by default)
    :return: True if this succeeds, False otherwise
    """
    for path in paths:
        pb_packet = reader.read_file(path)

        if out_dir is not None:
            file_name: str = os.path.basename(path).replace(".rdvxz", ".json")
            new_path = f"{out_dir}/{file_name}"
        else:
            new_path = path.replace(".rdvxz", ".json")

        with open(new_path, "w") as fout:
            fout.write(reader_utils.to_json(pb_packet))

        log.info("Converted %s to %s", path, new_path)

    return True


def rdvxm_to_json(paths: List[str], out_dir: Optional[str] = None) -> bool:
    """
    Converts .rdvxm files to .json files.
    :param paths: Paths of .rdvxm files to convert.
    :param out_dir: An optional output directory (will use input directory by default)
    :return: True if this succeeds, False otherwise
    """
    out_dir = out_dir if out_dir is not None else "."
    for path in paths:
        wrapped_packet: WrappedRedvoxPacketM = WrappedRedvoxPacketM.from_compressed_path(path)

        wrapped_packet.write_json_to_file(out_dir)
        log.info("Converted %s to json", path)

    return True


def json_to_rdvxz(paths: List[str], out_dir: Optional[str] = None) -> bool:
    """
    Converts .json files to .rdvxz files.
    :param paths: Paths of .json files to convert.
    :param out_dir: An optional output directory (will use input directory by default)
    :return: True if this succeeds, False otherwise
    """
    for path in paths:
        with open(path, "r") as fin:
            json: str = fin.read()

            if out_dir is not None:
                file_name: str = os.path.basename(path).replace(".json", ".rdvxz")
                new_path = f"{out_dir}/{file_name}"
            else:
                new_path = path.replace(".json", ".rdvxz")

            reader_utils.write_file(new_path, reader_utils.from_json(json))

            log.info("Converted %s to %s", path, new_path)

    return True


def json_to_rdvxm(paths: List[str], out_dir: Optional[str] = None) -> bool:
    """
    Converts .json files to .rdvxm files.
    :param paths: Paths of .json files to convert.
    :param out_dir: An optional output directory (will use input directory by default)
    :return: True if this succeeds, False otherwise
    """
    out_dir = out_dir if out_dir is not None else "."
    for path in paths:
        wrapped_packet: WrappedRedvoxPacketM = WrappedRedvoxPacketM.from_json_path(path)
        wrapped_packet.write_compressed_to_file(out_dir)
        log.info("Converted %s to .rdvxm", path)

    return True


def rdvxz_to_rdvxm(paths: List[str], out_dir: Optional[str] = None) -> bool:
    """
    Convert rdvxz files to rdvxm files
    :param paths: Paths of original files to convert
    :param out_dir: Optional output directory of converted files (default "./")
    :return: True if completed successfully
    """
    out_dir = out_dir if out_dir is not None else "."
    for path in paths:
        wrapped_packet_900: reader.WrappedRedvoxPacket = reader.read_rdvxz_file(path)
        wrapped_packet_1000: WrappedRedvoxPacketM = api_conversions.convert_api_900_to_1000(wrapped_packet_900)
        wrapped_packet_1000.write_compressed_to_file(out_dir)

    return True


def rdvxm_to_rdvxz(paths: List[str], out_dir: Optional[str] = None) -> bool:
    """
    Convert rdvxm files to rdvxz files
    :param paths: Paths of original files to convert
    :param out_dir: Optional output directory of converted files (default "./")
    :return: True if completed successfully
    """
    out_dir = out_dir if out_dir is not None else "."
    for path in paths:
        wrapped_packet_1000: WrappedRedvoxPacketM = WrappedRedvoxPacketM.from_compressed_path(path)
        wrapped_packet_900: reader.WrappedRedvoxPacket = api_conversions.convert_api_1000_to_900(wrapped_packet_1000)
        wrapped_packet_900.write_rdvxz(out_dir)

    return True


def rdvxz_print_stdout(paths: List[str]) -> bool:
    """
    Prints the contends of rdvxz files to sdtout.
    :param paths: Paths to .rdvxz files to print.
    :return: True if this succeeds, False otherwise.
    """
    for path in paths:
        print(reader.read_file(path))

    return True


def rdvxm_print_stdout(paths: List[str]) -> bool:
    """
    Prints the contends of rdvxm files to sdtout.
    :param paths: Paths to .rdvxm files to print.
    :return: True if this succeeds, False otherwise.
    """
    for path in paths:
        print(WrappedRedvoxPacketM.from_compressed_path(path))

    return True



from typing import Optional

from PySide2.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QTreeWidget

from redvox.api1000.gui.data_explorer.file_explorer import FileExplorer
from redvox.api1000.gui.data_explorer.packet_details import PlotAndDetailsColumn
from redvox.api1000.gui.data_explorer.packet_explorer import PacketExplorer


class FileAndPacketExplorerColumn(QWidget):
    def __init__(self, base_dir: str, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.file_explorer: FileExplorer = FileExplorer(base_dir, self)
        self.packet_explorer: PacketExplorer = PacketExplorer(self)

        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.file_explorer)
        self.layout().addWidget(self.packet_explorer)


class DataExplorer(QWidget):
    def __init__(self, base_dir: str, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.explorer_column = FileAndPacketExplorerColumn(base_dir, self)
        self.details_column = PlotAndDetailsColumn(self)

        self.setLayout(QHBoxLayout())
        self.layout().addWidget(self.explorer_column)
        self.layout().addWidget(self.details_column)
